#ident "$Id: zalloc.c,v 1.1 2001/10/20 02:12:22 hpa Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2000-2001 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, Inc.,
 *   59 Temple Place Ste 330, Boston MA 02111-1307, USA, version 2.1,
 *   incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/*
 * zalloc.c
 *
 * Allocate zero-initialized memory.  Try to be efficient if the area is
 * large.
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include "lpsm.h"
#include "internals.h"

void *lpsm_zalloc(size_t size)
{
  void *ptr, *mv;
  int xbit, obit, rorder;
  int order = BUDDY_ORDER_MIN;

  if ( size > AH->slab[0].size ) {
    while ( size > ((size_t)1 << order) )
      order++;

    xbit = lpsm_malloc_buddy_xbit(size);
    if ( !xbit ) {
      ptr = NULL;
    } else {
      rorder = AH->arena_size_lg2 - order;
      obit = 1 << rorder;
      ptr = (void *)((char *)AH->data_start +
		     ((uintptr_t)(xbit-obit) << order)); 
      if ( (uintptr_t)ptr & (PM->pagesize-1) ||
	   order < PM->pageshift ) {
	memset(ptr, 0, size);	/* Misaligned or too small */
      } else {
	size = (size_t)1 << order; /* "Real" size */
	/* Do an anonymous mmap() over the affected area */
	mv = mmap(ptr, size, PROT_READ|PROT_WRITE,
		  MAP_FIXED|MAP_PRIVATE, 0, 0);
	if ( mv != ptr ) {
	  memset(ptr, 0, size);
	} else {
	  /* Mark dirty */
	  memset(PM->pageinfo + ((ptr-PM->arena) >> PM->pageshift),
		 page_dirty, size >> PM->pageshift);
	}
      }
    }
  } else {
    ptr = lpsm_malloc_slab(size);
    if ( ptr )
      memset(ptr, 0, size);
  }

  return ptr;
}

